/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.clickhouse.reader.model;

import io.vertx.ext.sql.SQLRowStream;
import io.vertx.ext.web.RoutingContext;
import lombok.Getter;

import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicLong;
import java.util.concurrent.atomic.AtomicReference;

@Getter
public class ReaderProcessContext {
    private final AtomicReference<ReaderProcessState> state = new AtomicReference<>(new ReaderProcessState(0, null));
    private final AtomicLong uploadingCount = new AtomicLong(0);
    private final ConcurrentLinkedQueue<Chunk> chunksToUpload = new ConcurrentLinkedQueue<>();
    private final ConcurrentLinkedQueue<SQLRowStream> streams = new ConcurrentLinkedQueue<>();
    private final AtomicBoolean pausedStreams = new AtomicBoolean(false);
    private final RoutingContext routingContext;
    private final QueryRequest queryRequest;
    private final String startAddress;
    private final String processChunkAddress;

    public ReaderProcessContext(RoutingContext routingContext, QueryRequest queryRequest) {
        this.routingContext = routingContext;
        this.queryRequest = queryRequest;
        this.startAddress = queryRequest.getRequestId() + queryRequest.getTable() + queryRequest.getKafkaTopic() + "_START";
        this.processChunkAddress = queryRequest.getRequestId() + queryRequest.getTable() + queryRequest.getKafkaTopic() + "_PROCESS_CHUNK";
    }

    public ReaderProcessState getState() {
        return state.get();
    }

    public void updateStateFinishOneExecution() {
        state.updateAndGet(original -> new ReaderProcessState(original.getCompletedExecutions() + 1, original.getError()));
    }

    public void updateStateSetError(Throwable t) {
        state.updateAndGet(original -> {
            if (original.getError() != null) {
                return original;
            }
            return new ReaderProcessState(original.getCompletedExecutions(), t);
        });
    }

    public long getUploadingCount() {
        return uploadingCount.get();
    }

    public void incrementUploadingCount() {
        uploadingCount.incrementAndGet();
    }

    public void decrementUploadingCount() {
        uploadingCount.decrementAndGet();
    }
}
