/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.clickhouse.reader.service;

import io.vertx.core.CompositeFuture;
import io.vertx.core.Future;
import io.vertx.core.Vertx;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import org.springframework.stereotype.Service;
import ru.datamart.kafka.clickhouse.reader.configuration.properties.ReaderProperties;
import ru.datamart.kafka.clickhouse.reader.configuration.properties.VerticleProperties;
import ru.datamart.kafka.clickhouse.reader.factory.QueryResultUpstreamFactory;
import ru.datamart.kafka.clickhouse.reader.model.ReaderProcessContext;
import ru.datamart.kafka.clickhouse.reader.verticle.ClickhouseExecutorVerticle;
import ru.datamart.kafka.clickhouse.reader.verticle.ConfigurableVerticle;
import ru.datamart.kafka.clickhouse.reader.verticle.KafkaProduceVerticle;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.stream.Collectors;

@Service
@Slf4j
public class ReaderExecutionService {
    private final Random random = new Random();
    private final List<ClickhouseExecutor> databaseExecutors;
    private final Vertx vertx;
    private final QueryResultUpstreamFactory queryResultUpstreamFactory;
    private final VerticleProperties verticleProperties;
    private final ReaderProperties readerProperties;

    public ReaderExecutionService(List<ClickhouseExecutor> databaseExecutors,
                                  Vertx vertx,
                                  QueryResultUpstreamFactory queryResultUpstreamFactory,
                                  VerticleProperties verticleProperties,
                                  ReaderProperties readerProperties) {
        this.databaseExecutors = databaseExecutors;
        this.vertx = vertx;
        this.queryResultUpstreamFactory = queryResultUpstreamFactory;
        this.verticleProperties = verticleProperties;
        this.readerProperties = readerProperties;
    }

    public Future<Void> execute(ReaderProcessContext readerContext) {
        return prepareVertices(readerContext)
                .compose(deployedVertices -> runProcess(readerContext, deployedVertices));
    }

    private CompositeFuture prepareVertices(ReaderProcessContext readerProcessContext) {
        val verticesToDeploy = new ArrayList<ConfigurableVerticle>();

        if (readerProcessContext.getQueryRequest().isParallelMode()) {
            for (int i = 0; i < databaseExecutors.size(); i++) {
                val databaseExecutor = databaseExecutors.get(i);
                verticesToDeploy.add(new ClickhouseExecutorVerticle(readerProcessContext, databaseExecutor, readerProperties, verticleProperties.getExecutor(), i, databaseExecutors.size()));
            }
        } else {
            val randomDatabaseExecutor = databaseExecutors.get(random.nextInt(databaseExecutors.size()));
            verticesToDeploy.add(new ClickhouseExecutorVerticle(readerProcessContext, randomDatabaseExecutor, readerProperties, verticleProperties.getExecutor(), 0, 1));
        }


        for (int i = 0; i < verticleProperties.getProducer().getInstances(); i++) {
            verticesToDeploy.add(new KafkaProduceVerticle(readerProcessContext, queryResultUpstreamFactory, verticleProperties.getProducer()));
        }

        val successfullyDeployed = new ConcurrentLinkedQueue<String>();
        return CompositeFuture.join(verticesToDeploy.stream()
                        .map(verticle -> vertx.deployVerticle(verticle, verticle.getDeploymentOptions())
                                .onSuccess(deploymentId -> {
                                    successfullyDeployed.add(deploymentId);
                                    log.info("Deployed [{}]-[{}]", verticle.getClass().getSimpleName(), deploymentId);
                                }))
                        .collect(Collectors.toList()))
                .onFailure(t -> successfullyDeployed.forEach(deploymentId -> vertx.undeploy(deploymentId)
                        .onSuccess(ignored -> log.info("Undeployed [{}]", deploymentId))));
    }


    private Future<Void> runProcess(ReaderProcessContext readerProcessContext, CompositeFuture event) {
        return Future.future(promise -> {
            val vertices = event.<String>list();
            vertx.eventBus().publish(readerProcessContext.getStartAddress(), null);
            vertx.setPeriodic(100L, timerId -> {
                val state = readerProcessContext.getState();

                int expectedFinished = readerProcessContext.getQueryRequest().isParallelMode() ? databaseExecutors.size() : 1;
                if (state.getError() != null ||
                        state.getCompletedExecutions() == expectedFinished && readerProcessContext.getUploadingCount() == 0L) {
                    vertx.cancelTimer(timerId);
                    vertices.forEach(deploymentID -> vertx.undeploy(deploymentID).onSuccess(ignored -> log.info("Undeployed [{}]", deploymentID)));

                    if (state.getError() != null) {
                        promise.fail(state.getError());
                        return;
                    }

                    promise.complete();
                }
            });
        });
    }

}
