/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.clickhouse.reader.verticle;

import io.vertx.core.DeploymentOptions;
import io.vertx.core.Future;
import lombok.extern.slf4j.Slf4j;
import lombok.val;
import ru.datamart.kafka.clickhouse.reader.factory.QueryResultUpstreamFactory;
import ru.datamart.kafka.clickhouse.reader.model.ReaderProcessContext;
import ru.datamart.kafka.clickhouse.reader.upstream.QueryResultUpstream;

import static ru.datamart.kafka.clickhouse.reader.configuration.properties.VerticleProperties.WorkerProperties;

@Slf4j
public class KafkaProduceVerticle extends ConfigurableVerticle {
    private final ReaderProcessContext processContext;
    private final QueryResultUpstreamFactory queryResultUpstreamFactory;
    private final WorkerProperties properties;
    private QueryResultUpstream upstream;

    public KafkaProduceVerticle(ReaderProcessContext processContext,
                                QueryResultUpstreamFactory queryResultUpstreamFactory,
                                WorkerProperties properties) {
        this.processContext = processContext;
        this.queryResultUpstreamFactory = queryResultUpstreamFactory;
        this.properties = properties;
    }

    @Override
    public DeploymentOptions getDeploymentOptions() {
        return new DeploymentOptions()
                .setWorker(true)
                .setWorkerPoolName("kafka-producer-pool")
                .setWorkerPoolSize(properties.getPoolSize());
    }

    @Override
    public void start() {
        upstream = queryResultUpstreamFactory.create(processContext.getQueryRequest().getAvroSchema(), processContext.getQueryRequest().getKafkaBrokers());
        vertx.eventBus().consumer(processContext.getProcessChunkAddress(), ignored -> processChunk()
                .onComplete(ar -> {
                    if (ar.failed()) {
                        processContext.updateStateSetError(ar.cause());
                    }
                    processContext.decrementUploadingCount();
                }));
    }

    @Override
    public void stop() {
        if (upstream != null) {
            upstream.close();
        }
    }

    private Future<Void> processChunk() {
        return Future.future(p -> {
            if (processContext.getState().getError() != null) {
                p.complete();
                return;
            }

            val chunk = processContext.getChunksToUpload().poll();
            if (chunk == null) {
                p.complete();
                return;
            }
            val queryRequest = processContext.getQueryRequest();
            val table = queryRequest.getTable();
            val kafkaTopic = queryRequest.getKafkaTopic();
            val queryResultItem = chunk.getQueryResultItem();
            upstream.push(queryResultItem, table, kafkaTopic, chunk.getStreamId(), chunk.getStreamTotal())
                    .onSuccess(v -> {
                        log.debug("Chunk [{}] for table [{}] was pushed successfully into topic [{}] isLast [{}], stream number [{}]",
                                queryResultItem.getChunkNumber(), table, kafkaTopic, queryResultItem.getIsLastChunk(), chunk.getStreamId());
                        p.complete();
                    })
                    .onFailure(t -> {
                        log.error("Error sending chunk [{}] for table [{}] into topic [{}] isLast [{}], stream number [{}]",
                                queryResultItem.getChunkNumber(), table, kafkaTopic, queryResultItem.getIsLastChunk(), chunk.getStreamId(), t);
                        p.fail(t);
                    });
        });
    }
}
