/*
 * Copyright © 2022 DATAMART LLC
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package ru.datamart.kafka.clickhouse.reader.service;

import io.vertx.core.AsyncResult;
import io.vertx.core.Future;
import io.vertx.core.Handler;
import io.vertx.core.json.JsonArray;
import io.vertx.ext.sql.SQLClient;
import io.vertx.ext.sql.SQLConnection;
import io.vertx.ext.sql.SQLRowStream;
import lombok.val;
import org.apache.avro.Schema;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.extension.ExtendWith;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.junit.jupiter.MockitoExtension;
import ru.datamart.kafka.clickhouse.reader.converter.ClickhouseFromSqlTypeConverter;
import ru.datamart.kafka.clickhouse.reader.model.DatabaseExecutorRequest;
import ru.datamart.kafka.clickhouse.reader.model.QueryResultItem;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedList;
import java.util.List;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.contains;
import static org.hamcrest.Matchers.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

@ExtendWith(MockitoExtension.class)
class ClickhouseExecutorTest {
    private static final String SQL = "sql";

    @Mock
    private SQLClient sqlClient;

    @Mock
    private ClickhouseFromSqlTypeConverter sqlTypeConverter;

    @InjectMocks
    private ClickhouseExecutor clickhouseExecutor;


    @Mock
    private SQLConnection connection;

    @Mock
    private JsonArray row;

    private final RowStreamStub streamStub = new RowStreamStub();

    @BeforeEach
    void setUp() {
        lenient().when(sqlClient.getConnection(any())).thenAnswer(invocation -> {
            Handler<AsyncResult<SQLConnection>> argument = invocation.getArgument(0);
            argument.handle(Future.succeededFuture(connection));
            return sqlClient;
        });
        lenient().when(connection.queryStream(anyString(), any())).thenAnswer(invocation -> {
            Handler<AsyncResult<SQLRowStream>> argument = invocation.getArgument(1);
            argument.handle(Future.succeededFuture(streamStub));
            return connection;
        });
        lenient().when(sqlTypeConverter.convert(any(), any())).thenAnswer(invocation -> invocation.getArgument(1));
        lenient().when(row.size()).thenReturn(3);
        lenient().when(row.getValue(anyInt())).thenAnswer(invocationOnMock -> invocationOnMock.getArgument(0));
    }

    @Test
    void shouldSuccessWhenChunkSizeIsPerMessage() {
        // arrange
        val streams = new LinkedList<SQLRowStream>();
        val messages = new ArrayList<QueryResultItem>();

        val queryRequest = DatabaseExecutorRequest.builder()
                .sql(SQL)
                .chunkSize(1)
                .columnTypes(Arrays.asList(Schema.Type.INT, Schema.Type.INT, Schema.Type.INT))
                .build();

        // act execute
        Future<Void> result = clickhouseExecutor.execute(queryRequest, streams, messages::add);

        // act handle x3
        assertFalse(result.isComplete());
        streamStub.rowHandler.handle(row);
        streamStub.rowHandler.handle(row);
        streamStub.rowHandler.handle(row);

        // act end handler
        assertFalse(result.isComplete());
        streamStub.endHandler.handle(null);

        // assert
        assertTrue(result.isComplete());
        if (result.failed()) {
            fail(result.cause());
        }
        assertTrue(result.succeeded());

        assertThat(streams, contains(sameInstance(streamStub)));
        assertThat(messages, contains(
                allOf(
                        hasProperty("chunkNumber", is(1)),
                        hasProperty("isLastChunk", is(false)),
                        hasProperty("dataSet", contains(contains(is(0), is(1), is(2))))
                ),
                allOf(
                        hasProperty("chunkNumber", is(2)),
                        hasProperty("isLastChunk", is(false)),
                        hasProperty("dataSet", contains(contains(is(0), is(1), is(2))))
                ),
                allOf(
                        hasProperty("chunkNumber", is(3)),
                        hasProperty("isLastChunk", is(false)),
                        hasProperty("dataSet", contains(contains(is(0), is(1), is(2))))
                ),
                allOf(
                        hasProperty("chunkNumber", is(4)),
                        hasProperty("isLastChunk", is(true)),
                        hasProperty("dataSet", empty())
                )
        ));
    }

    @Test
    void shouldSuccessWhenChunkedAndOneLeftBehind() {
        // arrange
        val messages = new ArrayList<QueryResultItem>();
        val streams = new LinkedList<SQLRowStream>();

        val queryRequest = DatabaseExecutorRequest.builder()
                .sql(SQL)
                .chunkSize(2)
                .columnTypes(Arrays.asList(Schema.Type.INT, Schema.Type.INT, Schema.Type.INT))
                .build();

        // act execute
        Future<Void> result = clickhouseExecutor.execute(queryRequest, streams, messages::add);

        // act handle x3
        assertFalse(result.isComplete());
        streamStub.rowHandler.handle(row);
        streamStub.rowHandler.handle(row);
        streamStub.rowHandler.handle(row);
        streamStub.rowHandler.handle(row);
        streamStub.rowHandler.handle(row);

        // act end handler
        assertFalse(result.isComplete());
        streamStub.endHandler.handle(null);

        // assert
        assertTrue(result.isComplete());
        if (result.failed()) {
            fail(result.cause());
        }
        assertTrue(result.succeeded());

        assertThat(streams, contains(sameInstance(streamStub)));
        assertThat(messages, contains(
                allOf(
                        hasProperty("chunkNumber", is(1)),
                        hasProperty("isLastChunk", is(false)),
                        hasProperty("dataSet", contains(
                                contains(is(0), is(1), is(2)),
                                contains(is(0), is(1), is(2))
                        ))
                ),
                allOf(
                        hasProperty("chunkNumber", is(2)),
                        hasProperty("isLastChunk", is(false)),
                        hasProperty("dataSet", contains(
                                contains(is(0), is(1), is(2)),
                                contains(is(0), is(1), is(2))
                        ))
                ),
                allOf(
                        hasProperty("chunkNumber", is(3)),
                        hasProperty("isLastChunk", is(true)),
                        hasProperty("dataSet", contains(
                                contains(is(0), is(1), is(2))
                        ))
                )
        ));
    }

    @Test
    void shouldFailWhenConnectionFailed() {
        // arrange
        val messages = new ArrayList<QueryResultItem>();
        val streams = new LinkedList<SQLRowStream>();

        val queryRequest = DatabaseExecutorRequest.builder()
                .sql(SQL)
                .chunkSize(2)
                .columnTypes(Arrays.asList(Schema.Type.INT, Schema.Type.INT, Schema.Type.INT))
                .build();

        reset(sqlClient);
        when(sqlClient.getConnection(any())).thenAnswer(invocation -> {
            Handler<AsyncResult<SQLConnection>> argument = invocation.getArgument(0);
            argument.handle(Future.failedFuture(new RuntimeException("Exception")));
            return sqlClient;
        });

        // act execute
        Future<Void> result = clickhouseExecutor.execute(queryRequest, streams, messages::add);

        // assert
        assertTrue(result.isComplete());
        if (result.succeeded()) {
            fail(new AssertionError("Unexpected success"));
        }
        assertTrue(result.failed());
        assertSame(RuntimeException.class, result.cause().getClass());
        assertTrue(streams.isEmpty());
        assertTrue(messages.isEmpty());
    }

    @Test
    void shouldFailWhenQueryStreamFailed() {
        // arrange
        val messages = new ArrayList<QueryResultItem>();
        val streams = new LinkedList<SQLRowStream>();

        val queryRequest = DatabaseExecutorRequest.builder()
                .sql(SQL)
                .chunkSize(2)
                .columnTypes(Arrays.asList(Schema.Type.INT, Schema.Type.INT, Schema.Type.INT))
                .build();

        reset(connection);
        when(connection.queryStream(anyString(), any())).thenAnswer(invocation -> {
            Handler<AsyncResult<SQLRowStream>> argument = invocation.getArgument(1);
            argument.handle(Future.failedFuture(new RuntimeException("Exception")));
            return connection;
        });

        // act execute
        Future<Void> result = clickhouseExecutor.execute(queryRequest, streams, messages::add);

        // assert
        assertTrue(result.isComplete());
        if (result.succeeded()) {
            fail(new AssertionError("Unexpected success"));
        }
        assertTrue(result.failed());
        assertSame(RuntimeException.class, result.cause().getClass());
        assertTrue(streams.isEmpty());
        assertTrue(messages.isEmpty());
    }

    @Test
    void shouldFailWhenFailAfterHandle() {
        // arrange
        val messages = new ArrayList<QueryResultItem>();
        val streams = new LinkedList<SQLRowStream>();

        val queryRequest = DatabaseExecutorRequest.builder()
                .sql(SQL)
                .chunkSize(2)
                .columnTypes(Arrays.asList(Schema.Type.INT, Schema.Type.INT, Schema.Type.INT))
                .build();

        // act execute
        Future<Void> result = clickhouseExecutor.execute(queryRequest, streams, messages::add);

        // act handle and fail
        assertFalse(result.isComplete());
        streamStub.rowHandler.handle(row);

        // act end handler
        assertFalse(result.isComplete());
        streamStub.exceptionHandler.handle(new RuntimeException("Exception"));

        // assert
        assertTrue(result.isComplete());
        if (result.succeeded()) {
            fail(new AssertionError("Unexpected success"));
        }

        assertTrue(result.failed());
        assertSame(RuntimeException.class, result.cause().getClass());
        assertThat(streams, contains(sameInstance(streamStub)));
        assertTrue(messages.isEmpty());
    }

    static class RowStreamStub implements SQLRowStream {
        Handler<JsonArray> rowHandler;
        Handler<Void> endHandler;
        Handler<Throwable> exceptionHandler;

        @Override
        public SQLRowStream exceptionHandler(Handler<Throwable> handler) {
            exceptionHandler = handler;
            return this;
        }


        @Override
        public SQLRowStream handler(Handler<JsonArray> handler) {
            this.rowHandler = handler;
            return this;
        }

        @Override
        public SQLRowStream pause() {
            return this;
        }

        @Override
        public SQLRowStream resume() {
            return this;
        }

        @Override
        public SQLRowStream endHandler(Handler<Void> endHandler) {
            this.endHandler = endHandler;
            return this;
        }

        @Override
        public int column(String name) {
            return 0;
        }

        @Override
        public List<String> columns() {
            return null;
        }

        @Override
        public SQLRowStream resultSetClosedHandler(Handler<Void> handler) {
            return null;
        }

        @Override
        public void moreResults() {

        }

        @Override
        public SQLRowStream fetch(long l) {
            return this;
        }

        @Override
        public void close() {
        }

        @Override
        public void close(Handler<AsyncResult<Void>> completionHandler) {

        }
    }
}
